//
//
//    Copyright 2009 TheSeven
//
//
//    This file is part of the Linux4Nano toolkit.
//
//    TheSeven's iBugger is free software: you can redistribute it and/or
//    modify it under the terms of the GNU General Public License as
//    published by the Free Software Foundation, either version 2 of the
//    License, or (at your option) any later version.
//
//    TheSeven's iBugger is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//    See the GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License along
//    with the Linux4Nano toolkit.  If not, see <http://www.gnu.org/licenses/>.
//
//


#include <toolkit.h>
#include <ftl.h>
#include <fat32.h>


uint32_t fat32_ok;
uint32_t fat32_startsector;
uint32_t fat32_secperclus;
uint32_t fat32_database;
uint32_t fat32_fatbase;
uint32_t fat32_fatsize;
uint32_t fat32_fatcount;
uint32_t fat32_sectorcount;
uint32_t fat32_clustercount;
uint32_t fat32_rootdirclus;
uint32_t fat32_buf1[0x200] __attribute__((aligned(16)));
uint32_t fat32_buf2[0x200] __attribute__((aligned(16)));


uint32_t fat32_get_root()
{
    return fat32_rootdirclus;
}

uint32_t fat32_get_clusterchain(uint32_t clusterchain, uint32_t maxsize, void* buffer)
{
    uint32_t i;
    for (i = 0; i < (maxsize >> 11); )
    {
        uint32_t sector = (clusterchain - 2) * fat32_secperclus + fat32_database;
        uint32_t count = fat32_secperclus;
        if (count + i > (maxsize >> 11)) count = (maxsize >> 11) - i;
        uint32_t fatsector = fat32_fatbase + (clusterchain >> 9);
        if (ftl_read(fatsector, 1, &((uint32_t*)buffer)[i << 9])) return 1;
        clusterchain = ((uint32_t*)buffer)[(i << 9) + (clusterchain & 0x1FF)];
        if (ftl_read(sector, count, &((uint32_t*)buffer)[i << 9])) return 1;
        i += count;
        if (clusterchain >= 0x0ffffff0) return 0;
    }
    return clusterchain;
}

uint32_t fat32_get_direntry(uint32_t clusterchain, const char* filename, uint32_t* filesize)
{
    uint32_t i, j;
    while (clusterchain > 1 && clusterchain < 0x0ffffff0)
    {
        uint32_t sector = (clusterchain - 2) * fat32_secperclus + fat32_database;
        for (j = 0; j < fat32_secperclus; j++)
        {
            if (ftl_read(sector + j, 1, fat32_buf1)) return 1;
            for (i = 0; i < 0x200; i += 8)
                if (((uint8_t*)fat32_buf1)[i << 2] == 0) return 0;
                else if (((uint8_t*)fat32_buf1)[i << 2] == 0xe5) continue;
                else if (memcmp(&fat32_buf1[i], filename, 11) == 0)
                {
                    *filesize = fat32_buf1[i + 7];
                    return (((uint16_t*)fat32_buf1)[(i << 1) + 0xA] << 16)
                         | ((uint16_t*)fat32_buf1)[(i << 1) + 0xD];
                }
        }
        uint32_t fatsector = fat32_fatbase + (clusterchain >> 9);
        if (ftl_read(fatsector, 1, fat32_buf1)) return 1;
        clusterchain = fat32_buf1[(i << 9) + (clusterchain & 0x1FF)];
    }
    return 0;
}

uint32_t fat32_delete_clusterchain(uint32_t clusterchain)
{
    while (1)
    {
        uint32_t fatsector = fat32_fatbase + (clusterchain >> 9);
        if (ftl_read(fatsector, 1, fat32_buf1)) return 1;
        clusterchain = fat32_buf1[clusterchain & 0x1FF];
        fat32_buf1[(clusterchain & 0x1FF)] = 0;
        if (ftl_write(fatsector, 1, fat32_buf1)) return 1;
        if (clusterchain >= 0x0ffffff0) return 0;
    }
}

uint32_t fat32_delete_direntry(uint32_t clusterchain, const char* filename)
{
    uint32_t i, j;
    while (clusterchain > 1 && clusterchain < 0x0ffffff0)
    {
        uint32_t sector = (clusterchain - 2) * fat32_secperclus + fat32_database;
        for (j = 0; j < fat32_secperclus; j++)
        {
            if (ftl_read(sector + j, 1, fat32_buf1)) return 1;
            for (i = 0; i < 0x200; i += 8)
                if (((uint8_t*)fat32_buf1)[i << 2] == 0) return 0;
                else if (((uint8_t*)fat32_buf1)[i << 2] == 0xe5) continue;
                else if (memcmp(&fat32_buf1[i], filename, 11) == 0)
                {
                    ((uint8_t*)fat32_buf1)[i << 2] = 0xe5;
                    if (ftl_write(sector + j, 1, fat32_buf1)) return 1;
                    return 0;
                }
        }
        uint32_t fatsector = fat32_fatbase + (clusterchain >> 9);
        if (ftl_read(fatsector, 1, fat32_buf1)) return 1;
        clusterchain = fat32_buf1[(i << 9) + (clusterchain & 0x1FF)];
    }
    return 0;
}

uint32_t fat32_store_stream(void* buffer, uint32_t size)
{
    uint32_t i;
    uint32_t clusterchain = 0;
    uint32_t scanidx = 2;
    uint32_t scansect = 0xffffffff;
    uint32_t lastidx;
    uint32_t lastsect = 0xffffffff;
    uint32_t dirty = 0;
    while (size)
    {
        while (scanidx < fat32_clustercount + 2)
        {
            if ((scanidx >> 9) != scansect)
            {
                scansect = scanidx >> 9;
                if (ftl_read(fat32_fatbase + scansect, 1, fat32_buf1)) return 0;
            }
            if (!fat32_buf1[scanidx & 0x1ff]) break;
            scanidx++;
        }
        if (scanidx >= fat32_clustercount + 2) return 0;
        if (!clusterchain) clusterchain = scanidx;
        else
        {
            fat32_buf2[lastidx & 0x1ff] = scanidx;
            dirty = 1;
        }
        lastidx = scanidx;
        if ((lastidx >> 9) != lastsect)
        {
            if (dirty)
                if (ftl_write(fat32_fatbase + lastsect, 1, fat32_buf2)) return 0;
            dirty = 0;
            lastsect = lastidx >> 9;
            memcpy(fat32_buf2, fat32_buf1, 0x800);
        }
        uint32_t sector = (scanidx - 2) * fat32_secperclus + fat32_database;
        uint32_t count = (size + 0x7ff) >> 11;
        if (count > fat32_secperclus) count = fat32_secperclus;
        if (ftl_write(sector, count, &((uint32_t*)buffer)[i << 9])) return 0;
        if ((count << 11) >= size)
        {
            fat32_buf2[lastidx & 0x1ff] = 0x0fffffff;
            if (ftl_write(fat32_fatbase + lastsect, 1, fat32_buf2)) return 0;
            break;
        }
        size -= count << 11;
        buffer = (void*)((uint32_t)buffer + (count << 11));
        scanidx++;
    }
    return clusterchain;
}

void fat32_set_direntry(uint32_t* ptr, const char* filename, uint32_t filechain,
                        uint32_t filesize, uint32_t flags)
{
    memcpy(ptr, filename, 11);
    ((uint16_t*)ptr)[0xa] = filechain >> 16;
    ((uint8_t*)ptr)[0xb] = flags;
    ((uint16_t*)ptr)[0xa] = filechain >> 16;
    ((uint16_t*)ptr)[0xd] = filechain & 0xffff;
    ptr[7] = filesize;
}

uint32_t fat32_store_direntry(uint32_t dirchain, const char* filename,
                              uint32_t filechain, uint32_t filesize, uint32_t flags)
{
    uint32_t i, j;
    uint32_t lastidx;
    while (dirchain > 1 && dirchain < 0x0ffffff0)
    {
        uint32_t sector = (dirchain - 2) * fat32_secperclus + fat32_database;
        for (j = 0; j < fat32_secperclus; j++)
        {
            if (ftl_read(sector + j, 1, fat32_buf1)) return 1;
            for (i = 0; i < 0x200; i += 8)
                if (((uint8_t*)fat32_buf1)[i << 2] == 0
                 || ((uint8_t*)fat32_buf1)[i << 2] == 0xe5)
                {
                    fat32_set_direntry(&fat32_buf1[i], filename, filechain, filesize, flags);
                    if (ftl_write(sector + j, 1, fat32_buf1)) return 1;
                    return 0;
                }
        }
        uint32_t fatsector = fat32_fatbase + (dirchain >> 9);
        if (ftl_read(fatsector, 1, fat32_buf1)) return 1;
        lastidx = dirchain;
        dirchain = fat32_buf1[(i << 9) + (dirchain & 0x1FF)];
    }
    uint32_t scanidx = 2;
    uint32_t scansect = 0xffffffff;
    while (scanidx < fat32_clustercount + 2)
    {
        if ((scanidx >> 9) != scansect)
        {
            scansect = scanidx >> 9;
            if (ftl_read(fat32_fatbase + scansect, 1, fat32_buf1)) return 1;
        }
        if (!fat32_buf1[scanidx & 0x1ff]) break;
        scanidx++;
    }
    if (scanidx >= fat32_clustercount + 2) return 1;
    fat32_buf1[scanidx & 0x1ff] = 0x0fffffff;
    if (ftl_write(fat32_fatbase + scansect, 1, fat32_buf1)) return 1;
    if (ftl_read(fat32_fatbase + (lastidx >> 9), 1, fat32_buf1)) return 1;
    fat32_buf1[lastidx & 0x1ff] = scanidx;
    if (ftl_write(fat32_fatbase + scansect, 1, fat32_buf1)) return 1;
    uint32_t sector = (scanidx - 2) * fat32_secperclus + fat32_database;
    for (i = 0; i < fat32_secperclus; i++)
    {
        memset(fat32_buf1, 0, 0x800);
        if (!i) fat32_set_direntry(fat32_buf1, filename, filechain, filesize, flags);
        if (ftl_write(sector + i, 1, fat32_buf1)) return 1;
    }
    return 0;
}

uint32_t fat32_create_dir(uint32_t parent, const char* dirname)
{
    uint32_t i;
    uint32_t scanidx = 2;
    uint32_t scansect = 0xffffffff;
    while (scanidx < fat32_clustercount + 2)
    {
        if ((scanidx >> 9) != scansect)
        {
            scansect = scanidx >> 9;
            if (ftl_read(fat32_fatbase + scansect, 1, fat32_buf1)) return 0;
        }
        if (!fat32_buf1[scanidx & 0x1ff]) break;
        scanidx++;
    }
    if (scanidx >= fat32_clustercount + 2) return 0;
    fat32_buf1[scanidx & 0x1ff] = 0x0fffffff;
    if (ftl_write(fat32_fatbase + scansect, 1, fat32_buf1)) return 0;
    fat32_store_direntry(parent, dirname, scanidx, 0, 0x10);
    uint32_t sector = (scanidx - 2) * fat32_secperclus + fat32_database;
    for (i = 0; i < fat32_secperclus; i++)
    {
        memset(fat32_buf1, 0, 0x800);
        if (!i)
        {
            fat32_set_direntry(fat32_buf1, ".          ", scanidx, 0, 0x10);
            if (parent == fat32_rootdirclus) parent = 0;
            fat32_set_direntry(&fat32_buf1[8], "..         ", parent, 0, 0x10);
        }
        if (ftl_write(sector + i, 1, fat32_buf1)) return 0;
    }
    return scanidx;
}

uint32_t fat32_read_file(const char* filename, uint32_t maxsize, void* buffer, uint32_t* filesize)
{
}

uint32_t fat32_resize_patchdirs(uint32_t clusterchain, uint32_t clustoffset)
{
    uint32_t i, j;
    uint32_t* buffer = (uint32_t*)malloc(0x800);
    while (clusterchain < 0x0ffffff0)
    {
        uint32_t sectorbase = (clusterchain - 2) * fat32_secperclus + fat32_database;
        for (i = 0; i < fat32_secperclus; i++)
        {
            if (ftl_read(sectorbase + i, 1, buffer))
            {
                free(buffer);
                return 1;
            }
            for (j = 0; j < 64; j++)
                if (!((uint8_t*)buffer)[i << 2])
                {
                    free(buffer);
                    return 0;
                }
                else if (((uint8_t*)buffer)[j << 5] == 0xe5) continue;
                else if (((uint8_t*)buffer)[(j << 5) + 11] & 8) continue;
                else
                {
                    uint32_t clust = (((uint16_t*)buffer)[(j << 4) + 0xa] << 16)
                                    | ((uint16_t*)buffer)[(j << 4) + 0xd];
                    if (clust > 1 && clust < 0xffffff0)
                    {
                        clust += clustoffset;
                        ((uint16_t*)buffer)[(j << 4) + 0xa] = clust >> 16;
                        ((uint16_t*)buffer)[(j << 4) + 0xd] = clust & 0xffff;
                        if ((((uint8_t*)buffer)[(j << 5) + 0xb] & 0x10)
                         && memcmp(&((uint8_t*)buffer)[j << 5], ".          ", 11)
                         && memcmp(&((uint8_t*)buffer)[j << 5], "..         ", 11))
                            if (fat32_resize_patchdirs(clust, clustoffset))
                            {
                                free(buffer);
                                return 1;
                            }
                    }
                }
            if (ftl_write(sectorbase + i, 1, buffer))
            {
                free(buffer);
                return 1;
            }
        }
        uint32_t fatsector = fat32_fatbase + (clusterchain >> 9);
        if (ftl_read(fatsector, 1, buffer))
        {
            free(buffer);
            return 1;
        }
        clusterchain = buffer[(i << 9) + (clusterchain & 0x1FF)];
    }
    free(buffer);
    return 0;
}

uint32_t fat32_resize_fulldisk()
{
    uint32_t i, j;
    uint32_t fatsectors = 1;
    uint32_t oldfatsectors = 0;
    uint32_t clustercount;
    uint32_t reserved;
    uint32_t totalsectors = ftl_get_sector_count();
    if (!fat32_ok)
    {
        fat32_secperclus = 4;
        fat32_rootdirclus = 2;
    }
    while (fatsectors != oldfatsectors)
    {
        oldfatsectors = fatsectors;
        if (!fat32_ok) reserved = 2;
        else reserved = (fat32_database - fatsectors - 2) % fat32_secperclus + 2;
        clustercount = (totalsectors - fatsectors - reserved) / fat32_secperclus;
        fatsectors = (clustercount + 513) >> 9;
    }
    uint32_t database = fatsectors + reserved;
    uint32_t clusoffset;
    if (!fat32_ok) clusoffset = 0;
    else clusoffset = (fat32_database - database) / fat32_secperclus;
    memset(fat32_buf1, 0, 0x800);
    if (fat32_ok)
        if (ftl_read(fat32_startsector, 1, fat32_buf2))
        {
            fat32_ok = 0;
            return 1;
        }
    memcpy(fat32_buf1, "\xeb\x58\x00MSWIN5.0\0\x08", 0xd);
    ((uint8_t*)fat32_buf1)[0xd] = fat32_secperclus;
    ((uint16_t*)fat32_buf1)[7] = reserved;
    memcpy(&((uint8_t*)fat32_buf1)[0x10], "\x01\0\0\0\0\xf8\0\0\x3f\0\xff", 0xb);
    fat32_buf1[8] = totalsectors;
    fat32_buf1[9] = fatsectors;
    fat32_buf1[0xb] = fat32_rootdirclus + clusoffset;
    ((uint16_t*)fat32_buf1)[0x18] = 1;
    ((uint8_t*)fat32_buf1)[0x40] = 0x80;
    ((uint8_t*)fat32_buf1)[0x42] = 0x29;
    if (!fat32_ok) memcpy(&((uint8_t*)fat32_buf1)[0x43], "\0\0\0\0iPod Nano  ", 0xf);
    else memcpy(&((uint8_t*)fat32_buf1)[0x43], &((uint8_t*)fat32_buf2)[0x43], 0xf);
    memcpy(&((uint8_t*)fat32_buf1)[0x52], "FAT32   ", 8);
    ((uint16_t*)fat32_buf1)[0xff] = 0xaa55;
    if (ftl_write(0, 1, fat32_buf1))
    {
        fat32_ok = 0;
        return 1;
    }
    if (fat32_ok)
    {
        if (ftl_read(fat32_startsector + ((uint16_t*)fat32_buf2)[0x18], 1, fat32_buf1))
        {
            fat32_ok = 0;
            return 1;
        }
        fat32_buf1[0x7a] += clustercount - fat32_clustercount;
    }
    else
    {
        memset(fat32_buf1, 0, 0x800);
        fat32_buf1[0] = 0x41615252;
        fat32_buf1[0x79] = 0x61417272;
        fat32_buf1[0x7a] = clustercount - 1;
        fat32_buf1[0x7b] = 2;
        fat32_buf1[0xff] = 0xaa550000;
    }
    if (ftl_write(1, 1, fat32_buf1))
    {
        fat32_ok = 0;
        return 1;
    }
    uint32_t cursect = 0;
    if (!fat32_ok)
    {
        for (i = 0; i < fatsectors; i++)
        {
            memset(fat32_buf1, 0, 0x800);
            if (!i) memcpy(fat32_buf1, "\xf8\xff\xff\x0f\xff\xff\xff\xff\xff\xff\xff\x0f", 12);
            if (ftl_write(reserved + i, 1, fat32_buf1)) return 1;
        }
    }
    else
    {
        for (i = 0; i < fatsectors; i++)
        {
            memset(fat32_buf1, 0, 0x800);
            for (j = 0; j < 512; j++)
            {
                if (!i && !j) fat32_buf1[j] = 0x0fffffff;
                else if (!i && j == 1) fat32_buf1[j] = 0xffffffff;
                else if (((i << 9) | j) < clusoffset + 2);
                else if (((i << 9) | j) >= clusoffset + fat32_clustercount + 2);
                else
                {
                    uint32_t oldclust = (((i << 9) | j) - clusoffset);
                    if (((oldclust >> 9) + fat32_fatbase) != cursect)
                    {
                        cursect = (oldclust >> 9) + fat32_fatbase;
                        if (ftl_read(cursect, 1, fat32_buf2))
                        {
                            fat32_ok = 0;
                            return 1;
                        }
                    }
                    fat32_buf1[j] = fat32_buf2[oldclust & 0x1ff];
                    if (fat32_buf1[j] > 1 && fat32_buf1[j] < 0xffffff0)
                        fat32_buf1[j] += clusoffset;
                }
            }
            if (ftl_write(reserved + i, 1, fat32_buf1))
            {
                fat32_ok = 0;
                return 1;
            }
        }
    }
    fat32_startsector = 0;
    fat32_database = database;
    fat32_fatbase = reserved;
    fat32_fatsize = fatsectors;
    fat32_fatcount = 1;
    fat32_sectorcount = totalsectors;
    fat32_clustercount = clustercount;
    fat32_rootdirclus = fat32_rootdirclus + clusoffset;
    if (!fat32_ok)
    {
        for (i = 0; i < fat32_secperclus; i++)
        {
            memset(fat32_buf1, 0, 0x800);
            if (!i) memcpy(fat32_buf1, "iPod Nano  \x08", 12);
            if (ftl_write(database + i, 1, fat32_buf1)) return 1;
        }
    }
    else if (fat32_resize_patchdirs(fat32_rootdirclus, clusoffset))
    {
        fat32_ok = 0;
        return 1;
    }
    fat32_ok = 1;
    return 0;
}

uint32_t fat32_get_partition_start()
{
    return fat32_startsector;
}

uint32_t fat32_init()
{
    uint32_t i;
    fat32_ok = 0;
    fat32_startsector = 0xFFFFFFFF;
    if (ftl_init()) return 1;

    if (ftl_read(0, 1, fat32_buf1)) return 1;

    if (*((uint16_t*)((uint32_t)fat32_buf1 + 0x1FE)) != 0xAA55)
    {
#ifdef FAT32_DEBUG
        debug_printf("FAT32: Bad MBR!\n");
#endif
        return 1;
    }

    for (i = 0x1C2; i < 0x200; i += 0x10)
        if (((uint8_t*)fat32_buf1)[i] == 0xB)
        {
            fat32_startsector = *((uint16_t*)((uint32_t)fat32_buf1 + i + 4))
                              | (*((uint16_t*)((uint32_t)fat32_buf1 + i + 6)) << 16);
            break;
        }

    if (fat32_startsector == 0xFFFFFFFF
     && *((uint16_t*)((uint32_t)fat32_buf1 + 0x52)) == 0x4146
     && *((uint8_t*)((uint32_t)fat32_buf1 + 0x54)) == 0x54)
        fat32_startsector = 0;

    if (fat32_startsector == 0xFFFFFFFF)
    {
#ifdef FAT32_DEBUG
        debug_printf("FAT32: Can not find partition!\n");
#endif
        return 1;
    }

#ifdef FAT32_DEBUG
        debug_printf("FAT32: Found partition at sector %d\n", fat32_startsector);
#endif

    if (ftl_read(fat32_startsector, 1, fat32_buf1)) return 1;

    if (*((uint16_t*)((uint32_t)fat32_buf1 + 0x1FE)) != 0xAA55)
    {
#ifdef FAT32_DEBUG
        debug_printf("FAT32: Bad boot sector!\n");
#endif
        return 1;
    }

    if (((uint8_t*)fat32_buf1)[0xB] != 0 || ((uint8_t*)fat32_buf1)[0xC] != 8)
    {
#ifdef FAT32_DEBUG
        debug_printf("FAT32: Bad sector size!\n");
#endif
        return 1;
    }

    fat32_secperclus = ((uint8_t*)fat32_buf1)[0xD];
    uint32_t reserved = ((uint16_t*)fat32_buf1)[0x7];
    fat32_fatcount = ((uint8_t*)fat32_buf1)[0x10];

    if (((uint8_t*)fat32_buf1)[0x11] != 0)
    {
#ifdef FAT32_DEBUG
        debug_printf("FAT32: FAT12/FAT16 volume!\n");
#endif
        return 1;
    }

    fat32_sectorcount = fat32_buf1[8];
    fat32_fatsize = fat32_buf1[9];

    if (((uint16_t*)fat32_buf1)[0x15] != 0)
    {
#ifdef FAT32_DEBUG
        debug_printf("FAT32: Bad FS version!\n");
#endif
        return 1;
    }

    fat32_rootdirclus = fat32_buf1[0xB];

    fat32_clustercount = (fat32_sectorcount - reserved
                        - fat32_fatcount * fat32_fatsize) / fat32_secperclus;

    fat32_fatbase = fat32_startsector + reserved;
    fat32_database = fat32_fatbase + fat32_fatcount * fat32_fatsize;

    fat32_ok = 1;
    return 0;
}
