//
//
//    Copyright 2009 TheSeven
//
//
//    This file is part of the Linux4Nano toolkit.
//
//    TheSeven's iBugger is free software: you can redistribute it and/or
//    modify it under the terms of the GNU General Public License as
//    published by the Free Software Foundation, either version 2 of the
//    License, or (at your option) any later version.
//
//    TheSeven's iBugger is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//    See the GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License along
//    with the Linux4Nano toolkit.  If not, see <http://www.gnu.org/licenses/>.
//
//


#include <toolkit.h>
#include <util.h>


void memcpy(void* destination, const void* source, uint32_t size)
{
    while (size > 0)
    {
        ((uint8_t*)destination)[0] = ((uint8_t*)source)[0];
        source = (void*)((uint32_t)source + 1);
        destination = (void*)((uint32_t)destination + 1);
        size--;
    }
}

void memset(void* destination, uint8_t value, uint32_t size)
{
    while (size > 0)
    {
        ((uint8_t*)destination)[0] = value;
        destination = (void*)((uint32_t)destination + 1);
        size--;
    }
}

uint32_t memcmp(const void* destination, const void* source, uint32_t size)
{
    while (size > 0)
    {
        if (((uint8_t*)destination)[0] != ((uint8_t*)source)[0]) return 1;
        source = (void*)((uint32_t)source + 1);
        destination = (void*)((uint32_t)destination + 1);
        size--;
    }
    return 0;
}

void hexdump(const void* buffer, uint32_t size)
{
    uint32_t x, y, z;
    for (x = 0; ; x++)
    {
        debug_printf(">> %8X / %8X:  ", buffer + x * 16, x * 16);
        for (y = 0; y < 4; y++)
        {
            for (z = 0; z < 4; z++)
            {
                debug_printf("%2X ", ((uint8_t*)buffer)[(x * 4 + y) * 4 + z]);
                if ((x * 4 + y) * 4 + z + 1 >= size) return (void)debug_printf("\n");
            }
            debug_printf(" ");
        }
        debug_printf("\n");
    }
}

void clean_dcache()
{
    asm volatile(
        "MOV R0, #0                \n\t"
        "clean_dcache_loop:        \n\t"
        "MCR p15, 0, R0,c7,c10,2   \n\t"
        "ADD R1, R0, #0x10         \n\t"
        "MCR p15, 0, R1,c7,c10,2   \n\t"
        "ADD R1, R1, #0x10         \n\t"
        "MCR p15, 0, R1,c7,c10,2   \n\t"
        "ADD R1, R1, #0x10         \n\t"
        "MCR p15, 0, R1,c7,c10,2   \n\t"
        "ADDS R0, R0, #0x04000000  \n\t"
        "BNE clean_dcache_loop     \n\t"
        "MCR p15, 0, R0,c7,c10,4   \n\t"
        "MOV PC, LR                \n\t"
    );
}

void invalidate_dcache()
{
    asm volatile(
        "MOV R0, #0                \n\t"
        "invalidate_dcache_loop:   \n\t"
        "MCR p15, 0, R0,c7,c14,2   \n\t"
        "ADD R1, R0, #0x10         \n\t"
        "MCR p15, 0, R1,c7,c14,2   \n\t"
        "ADD R1, R1, #0x10         \n\t"
        "MCR p15, 0, R1,c7,c14,2   \n\t"
        "ADD R1, R1, #0x10         \n\t"
        "MCR p15, 0, R1,c7,c14,2   \n\t"
        "ADDS R0, R0, #0x04000000  \n\t"
        "BNE invalidate_dcache_loop\n\t"
        "MCR p15, 0, R0,c7,c10,4   \n\t"
        "MOV PC, LR                \n\t"
    );
}
