//
//
//    Copyright 2009-2010 TheSeven
//
//
//    This file is part of the Linux4Nano toolkit.
//
//    TheSeven's iBugger is free software: you can redistribute it and/or
//    modify it under the terms of the GNU General Public License as
//    published by the Free Software Foundation, either version 2 of the
//    License, or (at your option) any later version.
//
//    TheSeven's iBugger is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//    See the GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License along
//    with the Linux4Nano toolkit.  If not, see <http://www.gnu.org/licenses/>.
//
//



#include <toolkit.h>
#include <timer.h>
#include <aes.h>
#include <ftl.h>
#include <fat32.h>


extern void rendertext(void* buffer, uint32_t fgcol, uint32_t bgcol, char* text);
extern void displaylcd(uint32_t startx, uint32_t endx, uint32_t starty,
                       uint32_t endy, void* data, uint32_t color);
extern void i2csend(uint32_t device, uint32_t address, void* data, uint32_t length);
extern void i2crecv(uint32_t device, uint32_t address, void* data, uint32_t length);
extern void i2csendbyte(uint32_t device, uint32_t address, uint32_t data);
extern uint32_t i2crecvbyte(uint32_t device, uint32_t address);
extern void reset() __attribute__((noreturn));
extern uint32_t ucl_nrv2e_decompress_8(const void* src, uint32_t len_src, void* dst, uint32_t* plen_dst);
extern uint32_t ucl_nrv2e_test_8(const void* src, uint32_t len_src, void* dst, uint32_t* plen_dst);
extern uint32_t scriptstart;
#define script (&scriptstart)
#define scriptb ((uint8_t*)(&scriptstart))


unsigned char lcdtext[30];
uint16_t lcdbuffer[0x5AC0];
uint32_t statusbar_y;
uint32_t statusbar_offset;
uint32_t statusbar_divisor;
uint32_t statusbar_multiplier;
uint32_t statusbar_current;

#define MALLOCSIZE 0xC00000
uint32_t malloc_firstfree;
uint32_t mallocmap[MALLOCSIZE >> 5] __attribute__((aligned(16)));
uint8_t mallocbuf[MALLOCSIZE] __attribute__((aligned(16)));

uint8_t norbuf[0x100000] __attribute__((aligned(16)));
#define norbufword ((uint32_t*)norbuf)

#define nor ((uint8_t*)0x24000000)
#define norword ((uint32_t*)0x24000000)


void* malloc(uint32_t size)
{
    uint32_t blocks = (size + 0x1f) >> 4;
    uint32_t page = malloc_firstfree;
    uint32_t first = 1;
    while (page < MALLOCSIZE)
    {
        if (!(mallocmap[page >> 5] & (1 << (page & 0x1f))))
        {
            if (first) malloc_firstfree = page;
            uint32_t needed = blocks - 1;
            while (needed && page < MALLOCSIZE)
            {
                page++;
                if (mallocmap[page >> 5] & (1 << (page & 0x1f))) break;
                needed--;
            }
            if (!needed) break;
            first = 0;
        }
        page++;
    }
    if (page >= MALLOCSIZE)
    {
        rendertext(&lcdbuffer[0x3231], 0, 0xffff, "Error: Out of memory!");
        displaylcd(0, 175, 0, 131, lcdbuffer, 0);
        while (1);
    }
    page -= blocks - 1;
    void* ptr = (void*)((uint32_t)mallocbuf + ((page + 1) << 4));
    *((uint32_t*)((uint32_t)ptr - 16)) = blocks;
    while (blocks)
    {
        mallocmap[page >> 5] |= 1 << (page & 0x1f);
        page++;
        blocks--;
    }
    if (first) malloc_firstfree = page;
    return ptr;
}


void realloc(void* ptr, uint32_t size)
{
    uint32_t blocks = (size + 0x1f) >> 4;
    uint32_t oldblocks = *((uint32_t*)((uint32_t)ptr - 16));
    if (blocks > oldblocks)
    {
        rendertext(&lcdbuffer[0x3231], 0, 0xffff, "Error: Trying to grow alloc!");
        displaylcd(0, 175, 0, 131, lcdbuffer, 0);
        while (1);
    }
    *((uint32_t*)((uint32_t)ptr - 16)) = blocks;
    uint32_t page = (((uint32_t)ptr - (uint32_t)mallocbuf) >> 4) - 1 + blocks;
    blocks = oldblocks - blocks;
    if (page < malloc_firstfree) malloc_firstfree = page;
    while (blocks)
    {
        mallocmap[page >> 5] &= ~(1 << (page & 0x1f));
        page++;
        blocks--;
    }
}


void free(void* ptr)
{
    uint32_t blocks = *((uint32_t*)((uint32_t)ptr - 16));
    uint32_t page = (((uint32_t)ptr - (uint32_t)mallocbuf) >> 4) - 1;
    if (page < malloc_firstfree) malloc_firstfree = page;
    while (blocks)
    {
        mallocmap[page >> 5] &= ~(1 << (page & 0x1f));
        page++;
        blocks--;
    }
}


uint32_t freeret(uint32_t rc, void* ptr)
{
    free(ptr);
    return rc;
}


uint32_t decryptfw(void* image, uint32_t offset)
{
    uint32_t size = ((uint32_t*)image)[5];
    if (size > 0x800000) return 0;
    aes_decrypt(((uint32_t*)image)[2], &((uint8_t*)image)[offset], size);
    memcpy(image, &((uint8_t*)image)[offset], size);
    return size;
}


uint32_t getfw(const char* filename, uint32_t* sector, uint32_t* size)
{
    uint32_t i;
    uint32_t* buffer = malloc(0x800);
    if (ftl_read(0, 1, buffer) != 0) return freeret(1, buffer);
    if (*((uint16_t*)((uint32_t)buffer + 0x1FE)) != 0xAA55) return freeret(1, buffer);
    uint32_t startsector = 0;
    for (i = 0x1C2; i < 0x200; i += 0x10)
        if (((uint8_t*)buffer)[i] == 0)
        {
            startsector = *((uint16_t*)((uint32_t)buffer + i + 4))
                        | (*((uint16_t*)((uint32_t)buffer + i + 6)) << 16);
            break;
        }
    if (startsector == 0) return freeret(1, buffer);
    if (ftl_read(startsector, 1, buffer) != 0) return freeret(1, buffer);
    if (buffer[0x40] != 0x5B68695D) return freeret(1, buffer);
    if (ftl_read(startsector + 1 + (buffer[0x41] >> 11), 1, buffer) != 0)
        return freeret(1, buffer);
    for (i = 0; i < 0x1FE; i += 10)
        if (memcmp(&buffer[i], filename, 8) == 0)
        {
            *sector = startsector + (buffer[i + 3] >> 11);
            *size = buffer[i + 4] + 0x800;
            free(buffer);
            return 0;
        }
    return freeret(2, buffer);
}


uint32_t readfwptr(const char* filename, void* address, uint32_t* size)
{
    uint32_t sector;
    uint32_t rc = getfw(filename, &sector, size);
    if (rc) return rc;
    if (ftl_read(sector, ((*size + 0x7FF) >> 11), address) != 0) return 1;
    *size = decryptfw(address, 0x800);
    return 0;
}


uint32_t readfw(const char* filename, void** address, uint32_t* size)
{
    uint32_t sector;
    uint32_t rc = getfw(filename, &sector, size);
    if (rc) return rc;
    *address = malloc(*size);
    if (ftl_read(sector, ((*size + 0x7FF) >> 11), *address) != 0) return freeret(1, *address);
    *size = decryptfw(*address, 0x800);
    realloc(*address, *size);
    return 0;
}


uint32_t getnor(const char* filename, void** address, uint32_t* size)
{
    uint32_t i;
    for (i = 0; i < 0x1000; i += 0x10)
        if (memcmp(&nor[i], filename, 8) == 0)
        {
            *size = norword[(i + 0xc) >> 2];
            *address = &nor[norword[(i + 0x8) >> 2]];
            return 0;
        }
    return 1;
}


uint32_t readnorptr(const char* filename, void* address, uint32_t* size,
                    void* ucladdress, uint32_t* uclsize)
{
    void* noraddr;
    uint32_t rc = getnor(filename, &noraddr, size);
    if (rc) return rc;
    if (*size & 0x80000000)
    {
        *uclsize = (*size) & 0x7fffffff;
        if (ucladdress) memcpy(ucladdress, noraddr, *uclsize);
        if (address) return ucl_nrv2e_decompress_8(noraddr, *uclsize, address, size);
    }
    if (address) memcpy(address, noraddr, *size);
    if (ucladdress)
        return ucl_nrv2e_99_compress(noraddr, *size, ucladdress, uclsize, 0, 10, 0, 0);
    return 0;
}


uint32_t readnor(const char* filename, void** address, uint32_t* size,
                 void** ucladdress, uint32_t* uclsize)
{
    void* noraddr;
    uint32_t rc = getnor(filename, &noraddr, size);
    if (rc) return rc;
    if (*size & 0x80000000)
    {
        *uclsize = (*size) & 0x7fffffff;
        if (ucladdress)
        {
            *ucladdress = malloc(*uclsize);
            memcpy(*ucladdress, noraddr, *uclsize);
        }
        if (address)
        {
            rc = ucl_nrv2e_test_8(noraddr, *uclsize, *address, size);
            if (!rc)
            {
                *address = malloc(*size);
                rc = ucl_nrv2e_decompress_8(noraddr, *uclsize, *address, size);
            }
            if (rc)
            {
                if (ucladdress) free(*ucladdress);
                return freeret(rc, *address);
            }
            return 0;
        }
    }
    if (address)
    {
        *address = malloc(*size);
        memcpy(*address, noraddr, *size);
    }
    if (ucladdress)
    {
        *ucladdress = malloc(*size + (*size >> 3) + 256);
        rc = ucl_nrv2e_99_compress(noraddr, *size, *ucladdress, uclsize, 0, 10, 0, 0);
        if (rc)
        {
            if (address) free(*address);
            return freeret(rc, *ucladdress);
        }
        realloc(*ucladdress, *uclsize);
        return 0;
    }
    return 0;
}


uint32_t getapplenor(const char* filename, void** address, uint32_t* size)
{
    uint32_t i;
    for (i = 0xffe00; i < 0x100000; i += 0x28)
        if (memcmp(&nor[i], filename, 8) == 0)
        {
            *size = norword[(i + 0x10) >> 2] + 0x200;
            *address = &nor[norword[(i + 0xc) >> 2]];
            return 0;
        }
    return 1;
}


uint32_t readapplenorptr(const char* filename, void* address, uint32_t* size,
                         void* ucladdress, uint32_t* uclsize)
{
    void* noraddr;
    void* decryptaddr;
    uint32_t rc = getapplenor(filename, &noraddr, size);
    if (rc) return rc;
    if (address) decryptaddr = address;
    else decryptaddr = malloc(*size);
    memcpy(decryptaddr, noraddr, *size);
    *size = decryptfw(decryptaddr, 0x200);
    if (ucladdress)
        return ucl_nrv2e_99_compress(decryptaddr, *size, ucladdress, uclsize, 0, 10, 0, 0);
    if (!address) free(decryptaddr);
    return 0;
}


uint32_t readapplenor(const char* filename, void** address, uint32_t* size,
                      void** ucladdress, uint32_t* uclsize)
{
    void* noraddr;
    void* decryptaddr;
    uint32_t rc = getapplenor(filename, &noraddr, size);
    if (rc) return rc;
    decryptaddr = malloc(*size);
    memcpy(decryptaddr, noraddr, *size);
    *size = decryptfw(decryptaddr, 0x200);
    realloc(decryptaddr, *size);
    if (ucladdress)
    {
        *ucladdress = malloc(*size + (*size >> 3) + 256);
        rc = ucl_nrv2e_99_compress(decryptaddr, *size, *ucladdress, uclsize, 0, 10, 0, 0);
        if (rc) return freeret(freeret(rc, *ucladdress), decryptaddr);
        realloc(*ucladdress, *uclsize);
    }
    if (address) *address = decryptaddr;
    else free(decryptaddr);
    return 0;
}


void statusbar_init(uint32_t row, uint32_t range)
{
    if (!range) return;
    statusbar_y = row;
    statusbar_divisor = (range + 127) >> 7;
    uint32_t width;
    if (statusbar_divisor < 2)
    {
        statusbar_divisor = 0;
        statusbar_multiplier = 128 / (range + 1);
        width = (range + 1) * statusbar_multiplier;
    }
    else width = (range + 1) / statusbar_divisor;
    statusbar_offset = (176 - width) >> 1;
    displaylcd(statusbar_offset - 1, statusbar_offset - 1, row, row + 7, 0x40000000, 0);
    displaylcd(statusbar_offset + width + 1, statusbar_offset + width + 1,
               row, row + 7, 0x40000000, 0);
    displaylcd(statusbar_offset - 1, statusbar_offset + width + 1, row, row, 0x40000000, 0);
    displaylcd(statusbar_offset - 1, statusbar_offset + width + 1,
               row + 7, row + 7, 0x40000000, 0);
    statusbar_current = -1;
}


void statusbar_update(uint32_t pos)
{
    uint32_t x;
    if (!statusbar_divisor) x = pos * statusbar_multiplier;
    else x = pos / statusbar_divisor;
    if (statusbar_current == x) return;
    displaylcd(statusbar_offset + statusbar_current + 1, statusbar_offset + x,
               statusbar_y + 1, statusbar_y + 6, 0x40000000, 0x18);
    statusbar_current = x;
}


void main(void)
{
    uint32_t i, j, k;
    uint32_t appleflash;
    void* diskflsh;
    uint32_t disksize;
    void* diagflsh;
    uint32_t diagsize;
    void* diskucl;
    uint32_t diskuclsize;
    void* diagucl;
    uint32_t diaguclsize;
    void* osos;
    uint32_t osossize;

    memset(mallocmap, 0, 0x10000);
    malloc_firstfree = 0;

    memset(lcdbuffer, 0xff, 0xb580);
    rendertext(&lcdbuffer[0xb1], 0, 0xffffffff, "iLoader installer v0.1.3");
    rendertext(&lcdbuffer[0x631], 0, 0xffffffff, "Copyright 2010 TheSeven");
    displaylcd(0, 175, 0, 131, lcdbuffer, 0);
    i2csendbyte(0xe6, 0x2b, 10);
    i2csendbyte(0xe6, 0x2a, 6);
    i2csendbyte(0xe6, 0x28, 10);
    i2csendbyte(0xe6, 0x29, 1);
    i2csendbyte(0xe6, 0x35, 0x15);
    i2csendbyte(0xe6, 0x36, 0x01);
    PWRCON &= ~0x10;
    if (ftl_init())
    {
        rendertext(&lcdbuffer[0xbb1], 0, 0xffffffff, "Error: Couldn't mount FTL!");
        displaylcd(0, 175, 0, 131, lcdbuffer, 0);
        while (1);
    }
    if (norword[0x400] == 0x53436667) appleflash = 0;
    else if (norword[0x1000] == 0x53436667) appleflash = 1;
    else
    {
        rendertext(&lcdbuffer[0xbb1], 0, 0xffffffff, "Error: Unknown flash format!");
        displaylcd(0, 175, 0, 131, lcdbuffer, 0);
        while (1);
    }
    if (appleflash) rendertext(&lcdbuffer[0xbb1], 0, 0xffffffff, "Detected apple flash format");
    else rendertext(&lcdbuffer[0xbb1], 0, 0xffffffff, "Detected iLoader flash format");
    memset(norbuf, 0xff, 0x100000);
    rendertext(&lcdbuffer[0x16b1], 0, 0xffff, "Loading firmware...");
    rendertext(&lcdbuffer[0x1c31], 0, 0xffff, "(this may take a while)");
    displaylcd(0, 175, 0, 131, lcdbuffer, 0);

    osossize = 0;
    if (appleflash)
    {
        memcpy(&norbuf[0x1000], &nor[0x4000], 0x1000);
        if (readapplenor("hslfksid", &diskflsh, &disksize, &diskucl, &diskuclsize))
            diskuclsize = disksize = 0;
        if (readapplenor("hslfgaid", &diagflsh, &diagsize, &diagucl, &diaguclsize))
            diaguclsize = diagsize = 0;
        if (readfw("DNANkbso", &osos, &osossize)) osossize = 0;
        if (osossize)
        {
            if (((uint8_t*)osos)[0x64d48] == 0x2b && ((uint8_t*)osos)[0x64d54] == 0x34)
            {
                ((uint8_t*)osos)[0x64d48] = 0x43;
                ((uint8_t*)osos)[0x64d54] = 0x52;
            }
            if (((uint8_t*)osos)[0x3acd8] == 0x01)
                ((uint8_t*)osos)[0x3acd8] = 0x00;
        }
    }
    else
    {
        memcpy(&norbuf[0x1000], &nor[0x1000], 0x1000);
        if (readnor("diskmode", &diskflsh, &disksize, &diskucl, &diskuclsize))
            diskuclsize = disksize = 0;
        if (readnor("diagmode", &diagflsh, &diagsize, &diagucl, &diaguclsize))
            diaguclsize = diagsize = 0;
    }

    rendertext(&lcdbuffer[0x16b1], 0, 0xffff, "Arranging flash images...");
    displaylcd(0, 175, 0, 131, lcdbuffer, 0);
    uint32_t sp = 0;
    uint32_t beginptr = 0x2000;
    uint32_t endptr = 0x100000;
    uint32_t dirptr = 0;
    while (script[sp])
    {
        uint32_t file = script[sp] & 0xff;
        uint32_t flags = (script[sp] >> 8) & 0xff;
        uint32_t align = (script[sp] >> 16) & 0xff;
        void* data;
        uint32_t size;
        sp++;
        switch (file)
        {
            case 1:
                data = diskflsh;
                size = disksize;
                break;
            case 2:
                data = diskucl;
                size = diskuclsize;
                flags |= 2;
                break;
            case 3:
                data = diagflsh;
                size = diagsize;
                break;
            case 4:
                data = diagucl;
                size = diaguclsize;
                flags |= 2;
                break;
            default:
                data = &scriptb[script[sp++]];
                size = script[sp++];
        }
        if (size)
        {
            if (align && !(flags & 1))
            {
                if ((align << 12) < beginptr)
                {
                    snprintf(lcdtext, 30, "Error: Align failed! (%02X)", align);
                    rendertext(&lcdbuffer[0x16b1], 0, 0xffff, lcdtext);
                    displaylcd(0, 175, 0, 131, lcdbuffer, 0);
                    while (1);
                }
                beginptr = align << 12;
            }
            if (endptr - beginptr < size)
            {
                rendertext(&lcdbuffer[0x16b1], 0, 0xffff, "Error: Flash is full!    ");
                displaylcd(0, 175, 0, 131, lcdbuffer, 0);
                while (1);
            }
            uint32_t storesize = size;
            if (flags & 2) storesize |= 0x80000000;
            if (flags & 1)
            {
                endptr -= ((size + 0xfff) & ~0xfff);
                memcpy(&norbuf[endptr], data, size);
                file = endptr;
            }
            else
            {
                memcpy(&norbuf[beginptr], data, size);
                file = beginptr;
                beginptr += ((size + 0xfff) & ~0xfff);
            }
            if (!(flags & 4))
            {
                if (dirptr >= 0x1000)
                {
                    rendertext(&lcdbuffer[0x16b1], 0, 0xffff, "Error: Directory is full!");
                    displaylcd(0, 175, 0, 131, lcdbuffer, 0);
                    while (1);
                }
                memcpy(&norbuf[dirptr], &script[sp], 8);
                norbufword[(dirptr >> 2) + 2] = file;
                norbufword[(dirptr >> 2) + 3] = storesize;
                dirptr += 0x10;
                sp += 2;
            }
        }
        else if (!(flags & 4)) sp += 2;
    }
    free(diskflsh);
    free(diskucl);
    free(diagflsh);
    free(diagucl);

    uint32_t fatok;
    if (fat32_init() || fat32_get_partition_start())
    {
        rendertext(&lcdbuffer[0x16b1], 0, 0xffff, "Repartitioning...        ");
        displaylcd(0, 175, 0, 131, lcdbuffer, 0);
        fatok = !fat32_resize_fulldisk();
        if (!fatok) fatok = !fat32_resize_fulldisk();
        if (!fatok)
        {
            rendertext(&lcdbuffer[0x16b1], 0, 0xffff, "Error: Could not mount FS!");
            displaylcd(0, 175, 0, 131, lcdbuffer, 0);
            while (1);
        }
    }
    else fatok = 1;
    uint32_t size;
    uint32_t updating = 0;
    if (fatok)
    {
        uint32_t dir = fat32_get_direntry(fat32_get_root(), "ILOADER    ", &size);
        if (dir > 1) updating = 1;
        if (dir > 1 && fat32_get_direntry(dir, "UPDATE~1IGN", &size) > 1) fatok = 0;
    }
    if (fatok)
    {
        rendertext(&lcdbuffer[0x16b1], 0, 0xffff, "Installing files...      ");
        displaylcd(0, 175, 0, 131, lcdbuffer, 0);
        if (updating) sp = script[sp + 1];
        else sp = (sp + 2) << 2;
        uint32_t dir = fat32_get_root();
        uint32_t file;
        void* data;
        while (scriptb[sp])
            switch (scriptb[sp])
            {
                case 1:
                    dir = fat32_get_root();
                    sp += 4;
                    break;
                case 2:
                    file = fat32_get_direntry(dir, &scriptb[sp + 1], &size);
                    if (file < 2) file = fat32_create_dir(dir, &scriptb[sp + 1]);
                    if (file < 2)
                    {
                        rendertext(&lcdbuffer[0x3231], 0, 0xffff, "Error: Can't create dir!");
                        displaylcd(0, 175, 0, 131, lcdbuffer, 0);
                        snprintf(lcdtext, 30, "(at %8X)", sp);
                        rendertext(&lcdbuffer[0x37b1], 0, 0xffff, lcdtext);
                        displaylcd(0, 175, 0, 131, lcdbuffer, 0);
                        scriptb[sp] = 0;
                        sleep(3000000);
                        break;
                    }
                    dir = file;
                    sp += 12;
                    break;
                case 3:
                case 4:
                    file = fat32_get_direntry(dir, &scriptb[sp + 1], &size);
                    if (file > 1)
                    {
                        if (scriptb[sp] == 3)
                        {
                            sp += 20;
                            break;
                        }
                        if (fat32_delete_direntry(dir, &scriptb[sp + 1]))
                        {
                            sp += 20;
                            break;
                        }
                        if (fat32_delete_clusterchain(file))
                        {
                            sp += 20;
                            break;
                        }
                    }
                    size = script[(sp >> 2) + 4];
                    file = fat32_store_stream(&scriptb[script[(sp >> 2) + 3]], size);
                    if (file > 1 || !size)
                        fat32_store_direntry(dir, &scriptb[sp + 1], file, size, 0);
                    sp += 20;
                    break;
                case 5:
                case 6:
                    if (fat32_get_direntry(dir, &scriptb[sp + 1], &size) > 1)
                    {
                        sp += 12;
                        break;
                    }
                    data = osos;
                    size = osossize;
                    if (scriptb[sp] == 5)
                    {
                        if (!appleflash)
                        {
                            sp += 12;
                            break;
                        }
                        data = nor;
                        size = 0x100000;
                    }
                    if (size)
                    {
                        uint32_t file = fat32_store_stream(data, size);
                        if (file > 1) fat32_store_direntry(dir, &scriptb[sp + 1], file, size, 0);
                    }
                    sp += 12;
                    break;
                default:
                    rendertext(&lcdbuffer[0x3231], 0, 0xffff, "Error: Bad script!");
                    displaylcd(0, 175, 0, 131, lcdbuffer, 0);
                    snprintf(lcdtext, 30, "(at %8X)", sp);
                    rendertext(&lcdbuffer[0x37b1], 0, 0xffff, lcdtext);
                    displaylcd(0, 175, 0, 131, lcdbuffer, 0);
                    scriptb[sp] = 0;
                    sleep(3000000);
            }
        rendertext(&lcdbuffer[0x16b1], 0, 0xffff, "Syncing FTL...     ");
        displaylcd(0, 175, 0, 131, lcdbuffer, 0);
        ftl_sync();
    }

    rendertext(&lcdbuffer[0x16b1], 0, 0xffff, "Flashing... (DO NOT RESET!!!)");
    rendertext(&lcdbuffer[0x1c31], 0, 0xffff, "                           ");
    displaylcd(0, 175, 0, 131, lcdbuffer, 0);
    statusbar_init(41, 0x100000);
    flashnor(nor, norbuf, 0x100000);
    rendertext(&lcdbuffer[0x16b1], 0, 0xffff, "Installation successful!     ");
    displaylcd(0, 175, 0, 131, lcdbuffer, 0);
    sleep(1000000);
}


void irqhandler(void) __attribute__((interrupt("IRQ")));
void irqhandler(void)
{
    uint32_t sources;
    sources = SRCPND;

    SRCPND = sources;
    INTPND = sources;
}
