//
//
//    Copyright 2009-2010 TheSeven
//
//
//    This file is part of the Linux4Nano toolkit.
//
//    TheSeven's iBugger is free software: you can redistribute it and/or
//    modify it under the terms of the GNU General Public License as
//    published by the Free Software Foundation, either version 2 of the
//    License, or (at your option) any later version.
//
//    TheSeven's iBugger is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//    See the GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License along
//    with the Linux4Nano toolkit.  If not, see <http://www.gnu.org/licenses/>.
//
//



#include <toolkit.h>
#include <nand.h>


uint16_t lcdbuffer[0x5AC0];
uint32_t statusbar_y;
uint32_t statusbar_offset;
uint32_t statusbar_divisor;
uint32_t statusbar_multiplier;
uint32_t statusbar_current;


void statusbar_init(uint32_t row, uint32_t range)
{
    if (!range) return;
    statusbar_y = row;
    statusbar_divisor = (range + 127) >> 7;
    uint32_t width;
    if (statusbar_divisor < 2)
    {
        statusbar_divisor = 0;
        statusbar_multiplier = 128 / (range + 1);
        width = (range + 1) * statusbar_multiplier;
    }
    else width = (range + 1) / statusbar_divisor;
    statusbar_offset = (176 - width) >> 1;
    displaylcd(statusbar_offset - 1, statusbar_offset - 1, row, row + 7, 0x40000000, 0);
    displaylcd(statusbar_offset + width + 1, statusbar_offset + width + 1,
               row, row + 7, 0x40000000, 0);
    displaylcd(statusbar_offset - 1, statusbar_offset + width + 1, row, row, 0x40000000, 0);
    displaylcd(statusbar_offset - 1, statusbar_offset + width + 1,
               row + 7, row + 7, 0x40000000, 0);
    statusbar_current = -1;
}


void statusbar_update(uint32_t pos)
{
    uint32_t x;
    if (!statusbar_divisor) x = pos * statusbar_multiplier;
    else x = pos / statusbar_divisor;
    if (statusbar_current == x) return;
    displaylcd(statusbar_offset + statusbar_current + 1, statusbar_offset + x,
               statusbar_y + 1, statusbar_y + 6, 0x40000000, 0x18);
    statusbar_current = x;
}


void main(void)
{
    uint32_t i, j;
    const struct nand_device_info_type* ftl_nand_type;

    memset(lcdbuffer, 0xff, 0xb580);
    rendertext(&lcdbuffer[0xb1], 0, 0xffffffff, "resetftl v0.1.0");
    rendertext(&lcdbuffer[0x631], 0, 0xffffffff, "Copyright 2010 TheSeven");

    nand_init();
    ftl_nand_type = nand_get_device_type(0);

    rendertext(&lcdbuffer[0x16b1], 0, 0xffffffff, "Wiping flash...");
    displaylcd(0, 175, 0, 131, lcdbuffer, 0);
    statusbar_init(41, (*ftl_nand_type).blocks);
    for (i = 0; i < (*ftl_nand_type).blocks; i++)
    {
        for (j = 0; j < 4; j++)
            if (nand_get_device_type(j))
                nand_block_erase(j, i);
        statusbar_update(i + 1);
    }
}


void irqhandler(void) __attribute__((interrupt("IRQ")));
void irqhandler(void)
{
    uint32_t sources;
    sources = SRCPND;

    SRCPND = sources;
    INTPND = sources;
}
