//
//
//    Copyright 2009-2010 TheSeven
//
//
//    This file is part of the Linux4Nano toolkit.
//
//    TheSeven's iBugger is free software: you can redistribute it and/or
//    modify it under the terms of the GNU General Public License as
//    published by the Free Software Foundation, either version 2 of the
//    License, or (at your option) any later version.
//
//    TheSeven's iBugger is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//    See the GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License along
//    with the Linux4Nano toolkit.  If not, see <http://www.gnu.org/licenses/>.
//
//



#include <toolkit.h>
#include <aes.h>
#include <ftl.h>
#include <timer.h>


extern void rendertext(void* buffer, uint32_t fgcol, uint32_t bgcol, char* text);
extern void displaylcd(uint32_t startx, uint32_t endx, uint32_t starty,
                       uint32_t endy, void* data, uint32_t color);
extern void i2csend(uint32_t device, uint32_t address, void* data, uint32_t length);
extern void i2crecv(uint32_t device, uint32_t address, void* data, uint32_t length);
extern void i2csendbyte(uint32_t device, uint32_t address, uint32_t data);
extern uint32_t i2crecvbyte(uint32_t device, uint32_t address);
extern void reset() __attribute__((noreturn));


unsigned char lcdtext[30];
uint16_t lcdbuffer[0x5AC0];
uint32_t statusbar_y;
uint32_t statusbar_offset;
uint32_t statusbar_divisor;
uint32_t statusbar_multiplier;
uint32_t statusbar_current;

#define MALLOCSIZE 0xC00000
uint32_t malloc_firstfree;
uint32_t mallocmap[MALLOCSIZE >> 5] __attribute__((aligned(16)));
uint8_t mallocbuf[MALLOCSIZE] __attribute__((aligned(16)));

uint8_t norbuf[0x100000] __attribute__((aligned(16)));
#define norbufword ((uint32_t*)norbuf)

#define nor ((uint8_t*)0x24000000)
#define norword ((uint32_t*)0x24000000)


void* malloc(uint32_t size)
{
    uint32_t blocks = (size + 0x1f) >> 4;
    uint32_t page = malloc_firstfree;
    uint32_t first = 1;
    while (page < MALLOCSIZE)
    {
        if (!(mallocmap[page >> 5] & (1 << (page & 0x1f))))
        {
            if (first) malloc_firstfree = page;
            uint32_t needed = blocks - 1;
            while (needed && page < MALLOCSIZE)
            {
                page++;
                if (mallocmap[page >> 5] & (1 << (page & 0x1f))) break;
                needed--;
            }
            if (!needed) break;
            first = 0;
        }
        page++;
    }
    if (page >= MALLOCSIZE)
    {
        rendertext(&lcdbuffer[0x3231], 0, 0xffff, "Error: Out of memory!");
        displaylcd(0, 175, 0, 131, lcdbuffer, 0);
        while (1);
    }
    page -= blocks - 1;
    void* ptr = (void*)((uint32_t)mallocbuf + ((page + 1) << 4));
    *((uint32_t*)((uint32_t)ptr - 16)) = blocks;
    while (blocks)
    {
        mallocmap[page >> 5] |= 1 << (page & 0x1f);
        page++;
        blocks--;
    }
    if (first) malloc_firstfree = page;
    return ptr;
}


void realloc(void* ptr, uint32_t size)
{
    uint32_t blocks = (size + 0x1f) >> 4;
    uint32_t oldblocks = *((uint32_t*)((uint32_t)ptr - 16));
    if (blocks > oldblocks)
    {
        rendertext(&lcdbuffer[0x3231], 0, 0xffff, "Error: Trying to grow alloc!");
        displaylcd(0, 175, 0, 131, lcdbuffer, 0);
        while (1);
    }
    *((uint32_t*)((uint32_t)ptr - 16)) = blocks;
    uint32_t page = (((uint32_t)ptr - (uint32_t)mallocbuf) >> 4) - 1 + blocks;
    blocks = oldblocks - blocks;
    if (page < malloc_firstfree) malloc_firstfree = page;
    while (blocks)
    {
        mallocmap[page >> 5] &= ~(1 << (page & 0x1f));
        page++;
        blocks--;
    }
}


void free(void* ptr)
{
    uint32_t blocks = *((uint32_t*)((uint32_t)ptr - 16));
    uint32_t page = (((uint32_t)ptr - (uint32_t)mallocbuf) >> 4) - 1;
    if (page < malloc_firstfree) malloc_firstfree = page;
    while (blocks)
    {
        mallocmap[page >> 5] &= ~(1 << (page & 0x1f));
        page++;
        blocks--;
    }
}


uint32_t freeret(uint32_t rc, void* ptr)
{
    free(ptr);
    return rc;
}


uint32_t decryptfw(void* image, uint32_t offset)
{
    uint32_t size = ((uint32_t*)image)[5];
    if (size > 0x800000) return 0;
    aes_decrypt(((uint32_t*)image)[2], &((uint8_t*)image)[offset], size);
    memcpy(image, &((uint8_t*)image)[offset], size);
    return size;
}


uint32_t getfw(const char* filename, uint32_t* sector, uint32_t* size)
{
    uint32_t i;
    uint32_t* buffer = malloc(0x800);
    if (ftl_read(0, 1, buffer) != 0) return freeret(1, buffer);
    if (*((uint16_t*)((uint32_t)buffer + 0x1FE)) != 0xAA55) return freeret(1, buffer);
    uint32_t startsector = 0;
    for (i = 0x1C2; i < 0x200; i += 0x10)
        if (((uint8_t*)buffer)[i] == 0)
        {
            startsector = *((uint16_t*)((uint32_t)buffer + i + 4))
                        | (*((uint16_t*)((uint32_t)buffer + i + 6)) << 16);
            break;
        }
    if (startsector == 0) return freeret(1, buffer);
    if (ftl_read(startsector, 1, buffer) != 0) return freeret(1, buffer);
    if (buffer[0x40] != 0x5B68695D) return freeret(1, buffer);
    if (ftl_read(startsector + 1 + (buffer[0x41] >> 11), 1, buffer) != 0)
        return freeret(1, buffer);
    for (i = 0; i < 0x1FE; i += 10)
        if (memcmp(&buffer[i], filename, 8) == 0)
        {
            *sector = startsector + (buffer[i + 3] >> 11);
            *size = buffer[i + 4] + 0x800;
            free(buffer);
            return 0;
        }
    return freeret(2, buffer);
}


uint32_t readfwptr(const char* filename, void* address, uint32_t* size)
{
    uint32_t sector;
    uint32_t rc = getfw(filename, &sector, size);
    if (rc) return rc;
    if (ftl_read(sector, ((*size + 0x7FF) >> 11), address) != 0) return 1;
    *size = decryptfw(address, 0x800);
    return 0;
}


uint32_t readfw(const char* filename, void** address, uint32_t* size)
{
    uint32_t sector;
    uint32_t rc = getfw(filename, &sector, size);
    if (rc) return rc;
    *address = malloc(*size);
    if (ftl_read(sector, ((*size + 0x7FF) >> 11), *address) != 0) return freeret(1, *address);
    *size = decryptfw(*address, 0x800);
    realloc(*address, *size);
    return 0;
}




void statusbar_init(uint32_t row, uint32_t range)
{
    if (!range) return;
    statusbar_y = row;
    statusbar_divisor = (range + 127) >> 7;
    uint32_t width;
    if (statusbar_divisor < 2)
    {
        statusbar_divisor = 0;
        statusbar_multiplier = 128 / (range + 1);
        width = (range + 1) * statusbar_multiplier;
    }
    else width = (range + 1) / statusbar_divisor;
    statusbar_offset = (176 - width) >> 1;
    displaylcd(statusbar_offset - 1, statusbar_offset - 1, row, row + 7, 0x40000000, 0);
    displaylcd(statusbar_offset + width + 1, statusbar_offset + width + 1,
               row, row + 7, 0x40000000, 0);
    displaylcd(statusbar_offset - 1, statusbar_offset + width + 1, row, row, 0x40000000, 0);
    displaylcd(statusbar_offset - 1, statusbar_offset + width + 1,
               row + 7, row + 7, 0x40000000, 0);
    statusbar_current = -1;
}


void statusbar_update(uint32_t pos)
{
    uint32_t x;
    if (!statusbar_divisor) x = pos * statusbar_multiplier;
    else x = pos / statusbar_divisor;
    if (statusbar_current == x) return;
    displaylcd(statusbar_offset + statusbar_current + 1, statusbar_offset + x,
               statusbar_y + 1, statusbar_y + 6, 0x40000000, 0x18);
    statusbar_current = x;
}


void main(void)
{
    uint32_t i, j, k;
    uint8_t* aupd;
    uint32_t aupdsize;
	uint32_t payloadstart = 0;

    memset(mallocmap, 0, 0x10000);
    malloc_firstfree = 0;

    memset(lcdbuffer, 0xff, 0xb580);
    rendertext(&lcdbuffer[0xb1], 0, 0xffffffff, "iLoader uninstaller v0.1.2");
    rendertext(&lcdbuffer[0x631], 0, 0xffffffff, "Copyright 2010 TheSeven");
    displaylcd(0, 175, 0, 131, lcdbuffer, 0);
    if (ftl_init())
    {
        rendertext(&lcdbuffer[0xbb1], 0, 0xffff, "Error: Couldn't mount FTL!");
        displaylcd(0, 175, 0, 131, lcdbuffer, 0);
        while (1);
    }
    if (norword[0x400] != 0x53436667)
    {
        rendertext(&lcdbuffer[0xbb1], 0, 0xffff, "Error: No SYSCFG!");
        displaylcd(0, 175, 0, 131, lcdbuffer, 0);
        while (1);
    }
    memset(norbuf, 0xff, 0x100000);
    rendertext(&lcdbuffer[0x16b1], 0, 0xffff, "Loading flash update...");
    displaylcd(0, 175, 0, 131, lcdbuffer, 0);

    memcpy(&norbuf[0x4000], &nor[0x1000], 0x1000);
    aupdsize = 0;
    if (readfw("DNANdpua", (void**)&aupd, &aupdsize)) aupdsize = 0;
    if (!aupdsize)
    {
        rendertext(&lcdbuffer[0x16b1], 0, 0xffff, "Error: Can't load AUPD!");
        displaylcd(0, 175, 0, 131, lcdbuffer, 0);
        while (1);
    }
	for (i = 0; i < (aupdsize >> 2); i++)
		if (((uint32_t*)aupd)[i] == 0x50796c64 && ((uint32_t*)aupd)[i + 4] == 0x46775570)
		    payloadstart = i << 2;
	if (!payloadstart)
    {
        rendertext(&lcdbuffer[0x16b1], 0, 0xffff, "Error: Can't find payload!");
        displaylcd(0, 175, 0, 131, lcdbuffer, 0);
        while (1);
    }
	memcpy(norbuf, &aupd[payloadstart + 0x2c], 0x2000);
	memcpy(&norbuf[0x8800], &aupd[payloadstart + 0x2048], 0x1f800);
	memcpy(&norbuf[0x28000], &aupd[payloadstart + 0x22048], 0xd8000);
	free(aupd);

    rendertext(&lcdbuffer[0x16b1], 0, 0xffff, "Encrypting flash update...");
    displaylcd(0, 175, 0, 131, lcdbuffer, 0);
    memset(&norbuf[0x8000], 0, 0x800);
    norbufword[0x2000] = 0x31303738;
    norbufword[0x2001] = 0x00302e31;
    norbufword[0x2002] = 0x800;
    norbufword[0x2003] = 0x1f800;
    hash(&norbuf[0x8800], 0x7e00, &norbuf[0x8010]);
    hash(&norbuf[0x8000], 0x10, &norbuf[0x8040]);
    aes_encrypt(2, &norbuf[0x8000], 0x20000);
    for (i = 0xffe00; i < 0xffec8; i += 0x28)
        if (norbufword[i >> 2])
        {
            uint32_t offs = norbufword[(i >> 2) + 3] >> 2;
            norbufword[offs] = 0;
            norbufword[offs + 1] = 2;
            norbufword[offs + 2] = 2;
            norbufword[offs + 3] = 0x40;
            norbufword[offs + 4] = 0;
            norbufword[offs + 5] = norbufword[(i >> 2) + 4];
            hash(&norbufword[offs + 0x80], norbufword[(i >> 2) + 4] >> 2, &norbufword[offs + 7]);
            memset(&norbufword[offs + 0x75], 0, 0x14);
            hash(&norbufword[offs], 0x80, &norbufword[offs + 0x75]);
            aes_encrypt(2, &norbufword[offs + 0x80], norbufword[(i >> 2) + 4]);
        }

    rendertext(&lcdbuffer[0x16b1], 0, 0xffff, "Flashing... (DO NOT RESET!!!)");
    displaylcd(0, 175, 0, 131, lcdbuffer, 0);
    statusbar_init(41, 0x100000);
    flashnor(nor, norbuf, 0x100000);
    rendertext(&lcdbuffer[0x16b1], 0, 0xffff, "Uninstallation successful!   ");
    displaylcd(0, 175, 0, 131, lcdbuffer, 0);
    sleepms(1000);
}


void irqhandler(void) __attribute__((interrupt("IRQ")));
void irqhandler(void)
{
}
